<?php

use WHMCS_DirectAdmin_SiteBuilder_Module\SiteproApiClient;
use WHMCS_DirectAdmin_SiteBuilder_Module\SiteproModule;

require_once __DIR__.'/SiteproApiClient.php';
require_once __DIR__.'/SiteproInternalApi.php';
require_once __DIR__.'/SiteproModule.php';

class SiteproBuilder {
	private $loginKeyFile;

	private static $instance;
	
	private $enterprise = true;

	private $packageFile = '/usr/local/directadmin/data/admin/custom_package_items.conf';

	/** @var string */
	private $domain;

	/** @var string */
	private $user;

	/**
	 * User home directory
	 * @var string|null
	 */
	private $homeDir = null;
	
	/**
	 * User web root directory for current domain
	 * @var string|null
	 */
	private $webRootDir = null;

	/** @var SiteproModule|null */
	private $module = null;

	/**
	 * @return SiteproBuilder
	 */
	public static function &getInstance() {
		if (!self::$instance) {
			self::$instance = new self();
		}
		return self::$instance;
	}

	public function __construct() {
		$this->loginKeyFile = dirname(__FILE__).'/login_key';
		$this->user = $this->resolveCurrentUser();
		$this->domain = $this->resolveCurrentDomain();
	}

	public function addPackage($name, array $params) {
		$this->removePackage($name);
		file_put_contents($this->packageFile, "\n".$name."=".http_build_query($params)."\n", FILE_APPEND);
	}

	public function removePackage($name) {
		$data = is_file($this->packageFile) ? explode("\n", file_get_contents($this->packageFile)) : array();
		$name_str = $name.'=';
		$name_len = strlen($name_str);
		if (($fh = @fopen($this->packageFile, 'w')) !== false) {
			foreach ($data as $line) {
				if ($line && substr($line, 0, $name_len) != $name_str) {
					fwrite($fh, $line."\n");
				}
			}
			fclose($fh);
		}
	}

	public function setLoginKey($key) {
		file_put_contents($this->loginKeyFile, $key);
	}
	
	public function getLoginKey() {
		return is_file($this->loginKeyFile) ? file_get_contents($this->loginKeyFile) : null;
	}

	/** @return string */
	public function getDomain() {
		return $this->domain;
	}

	/**
	 * @param string $domain
	 * @param string|null $subdomain
	 */
	public function setDomain($domain, $subdomain = null) {
		if ($subdomain) {
			$this->domain = substr($domain, strlen($subdomain) + 1);
		} else {
			$this->domain = $domain;
		}
	}

	private function resolveCurrentDomain() {
		return (isset($_GET['d']) && $_GET['d'] && $_GET['d'] != '|DOMAIN|') ? $_GET['d']
			: ((isset($_GET['domain']) && $_GET['domain'] && $_GET['domain'] != '|DOMAIN|') ? $_GET['domain']
			: ((isset($_SERVER['SESSION_SELECTED_DOMAIN']) && $_SERVER['SESSION_SELECTED_DOMAIN']) ? $_SERVER['SESSION_SELECTED_DOMAIN'] : null));
	}

	private function resolveSubdomain() {
		return (isset($_GET['sd']) && $_GET['sd']) ? $_GET['sd'] : null;
	}

	private function resolveCurrentUser() {
		return isset($_SERVER['USER']) ? $_SERVER['USER'] : 'demo';
	}

	public function setUser($user) {
		$this->user = $user;
	}

	private function resolvePort() {
		return isset($_SERVER['SERVER_PORT']) ? $_SERVER['SERVER_PORT'] : 2222;
	}

	private function resolveProto() {
		return (isset($_SERVER['SSL']) && $_SERVER['SSL'] || (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] && strtolower($_SERVER['HTTPS']) != 'off')) ? 'https' : 'http';
	}

	private function resolveServerAddr() {
		list($servAddr) = explode(':', isset($_SERVER['SERVER_ADDR']) ? $_SERVER['SERVER_ADDR'] : '', 2);
		return $servAddr;
	}

	private function resolveHost() {
		list($host) = explode(':', isset($_SERVER['HTTP_HOST']) ? $_SERVER['HTTP_HOST'] : '', 2);
		if (!$host || $host == 'localhost' || $host == '127.0.0.1' || function_exists('filter_var') && !filter_var($host, FILTER_VALIDATE_IP) && gethostbyname($host) == $host) {
			$host = self::resolveServerAddr();
		}
		return $host;
	}

	private function resolveBaseUri() {
		list($baseUri) = explode('?', (isset($_SERVER['REQUEST_URI']) ?  $_SERVER['REQUEST_URI'] : ''), 2);
		return $baseUri;
	}

	public function &getModule() {
		if (is_null($this->module)) {
			$host = $this->resolveHost();
			$port = $this->resolvePort();
			$proto = $this->resolveProto();
			$this->module = new SiteproModule();
			$this->module->setDomain($this->domain);
			$this->module->setUser($this->user);
			$this->module->setHost($host);
			$this->module->setPort($port);
			$this->module->setProto($proto);
			$this->module->setLoginKey($this->getLoginKey());
			$this->module->setBaseUri($this->resolveBaseUri());
			$this->module->setApi(new SiteproInternalApi($host, $this->user, null, $port, $proto));
		}
		return $this->module;
	}

	public function run() {
		try {
			$subdomain = $this->resolveSubdomain();
			if ($this->getArg(0) == 'api') {
				$this->handleApi($subdomain);
				exit();
			} else {
				if (!$this->getModule()->isEnabled())
					throw new ErrorException('This feature is not available for your account');
				$url = $this->getModule()->getBuilderUrl($subdomain);
				self::exitWithResponse('<script>window.top.location.href="'.$url.'";</script>');
			}
		} catch (ErrorException $ex) {
			self::exitWithResponse($ex->getMessage());
		}
	}

	public static function exitWithResponse($text, $asJson = false) {
		echo "HTTP/1.1 200 OK\r\n";
		echo "Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0\r\n";
		echo "Date: ".date('D, j M Y H:i:s')." GMT\r\n";
		echo "Expires: ".date('D, j M Y H:i:s')." GMT\r\n";
		echo "Connection: close\r\n";
		echo "Content-Length: ".strlen($text)."\r\n";
		echo "Content-type: ".($asJson ? "application/json" : "text/html")."; charset=utf-8\r\n";
		echo "\r\n";
		echo $text;
		exit();
	}
	
	public static function exitWithRedirect($url) {
		echo "HTTP/1.1 302 OK\r\n";
		echo "Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0\r\n";
		echo "Date: ".date('D, j M Y H:i:s')." GMT\r\n";
		echo "Expires: ".date('D, j M Y H:i:s')." GMT\r\n";
		echo "Connection: close\r\n";
		echo "Content-type: text/html; charset=utf-8\r\n";
		echo "Location: ".$url."\r\n";
		echo "\r\n";
		exit();
	}

	/** @var array|null */
	public $args = null;

	/** @return string[] */
	public function &getArgs() {
		if (is_null($this->args)) {
			$uriParts = explode('/boot.php/', (isset($_SERVER['REQUEST_URI']) ?  $_SERVER['REQUEST_URI'] : ''), 2);
			$rufx = count($uriParts) > 1 ? trim($uriParts[1], '/') : '';
			$this->args = explode('/', $rufx);
		}
		return $this->args;
	}
	
	public function getArg($idx) {
		$args = $this->getArgs();
		return isset($args[$idx]) ? $args[$idx] : null;
	}

	public function handleApi($subdomain = null) {
		@set_time_limit(3600);
		$gv = array();
		if (isset($_SERVER['QUERY_STRING'])) {
			$qs = str_replace('?', '&', $_SERVER['QUERY_STRING']);
			parse_str($qs, $gv);
			if (!is_array($gv)) $gv = array();
		}
		$post = (!empty($_POST)) ? $_POST : (isset($_SERVER['POST']) ? $_SERVER['POST'] : null);
		if (is_array($post) && !empty($post)) {
			$gv = array_merge($gv, $post);
		}
		if (isset($gv['X-Sitepro-Meta'])) {
			$data = $gv['X-Sitepro-Meta'];
			$gv['X-Sitepro-Meta'] = array();
			parse_str($data, $gv['X-Sitepro-Meta']);
		}
		$au = (isset($_SERVER['PHP_AUTH_USER']) && $_SERVER['PHP_AUTH_USER']) ? $_SERVER['PHP_AUTH_USER'] : null;
		$ap = (isset($_SERVER['PHP_AUTH_PW']) && $_SERVER['PHP_AUTH_PW']) ? $_SERVER['PHP_AUTH_PW'] : null;
		$domain = (isset($gv['domain']) && $gv['domain']) ? $gv['domain'] : ((isset($gv['X-Sitepro-Meta']['domain']) && $gv['X-Sitepro-Meta']['domain']) ? $gv['X-Sitepro-Meta']['domain'] : null);
		if (isset($gv['X-Sitepro-Meta']['user']) && isset($gv['X-Sitepro-Meta']['hash'])) {
			$au = $gv['X-Sitepro-Meta']['user'];
			$ap = $gv['X-Sitepro-Meta']['hash'];
		}
		$user = (object) array('username' => $au, 'domain' => $domain);
		$api = new SiteproApiClient();
		if ($au && $ap && (!$this->enterprise || ($user = $api->getUser($au, $ap, $domain)) && (!isset($user->error) || !$user->error))) {
			if (preg_match('#^admin\|(.+)$#', $user->username, $m)) {
				$user->username = $m[1];
			}
			$this->setUser($user->username);
			$this->setDomain($user->domain, $subdomain);
		} else if (isset($_SERVER['SESSION_ID']) && $_SERVER['SESSION_ID'] &&
			isset($_SERVER['SESSION_KEY']) && $_SERVER['SESSION_KEY'] &&
			$this->user && $this->domain) {
		} else {
			$err = (isset($user) && $user && isset($user->error) && $user->error)
					? $user->error->message : null;
			$arr = array(
					'result' => null,
					'error' => array(
						'message' => ($err ? $err : "API authentification error"),
						'code' => 401
					)
				);
			SiteproBuilder::exitWithResponse(json_encode($arr), true);
		}
		
		$method = isset($_SERVER['REQUEST_METHOD']) ? $_SERVER['REQUEST_METHOD'] : 'GET';
		if ($method == 'GET') {
			$args = $this->getArgs();
			$id = (count($args) > 1) ? intval($args[count($args) - 1]) : null;
			$finished = ($this->loadState() == 100);
			$this->saveState(null);
			$arr = array(
					'result' => array('id' => $id, 'status' => ($finished ? 'FINISHED' : 'FAILED'), 'output' => ($finished ? '' : 'Site downloading was interrupted.')),
					'error' => null
				);
			SiteproBuilder::exitWithResponse(json_encode($arr), true);
		} else if ($method == 'POST') {
			if ((!$_POST || !is_array($_POST) || empty($_POST)) && isset($_SERVER['POST'])) {
				parse_str($_SERVER['POST'], $_POST);
			}
			$account_id = isset($_POST['account_id']) ? $_POST['account_id'] : null;
			$zip = isset($_POST['zip']) ? $_POST['zip'] : null;
			$id = time();
			if ($zip) {
				ob_start();
				try {
					$this->saveState(0);
					$this->extractFromUrl($zip);
					$this->saveState(100);
				} catch (ErrorException $ex) {
					SiteproBuilder::exitWithResponse(json_encode(array('error' => $ex->getMessage())), true);
				}
				$buffOut = ob_get_clean();
			}
			$arr = array(
					'result' => array('id' => $id),
					'error' => null,
					'error2' => (isset($buffOut) ? $buffOut : null)
				);
			SiteproBuilder::exitWithResponse(json_encode($arr), true);
		}
		exit();
	}

	private function saveState($state) {
		$file = $this->getWebRootDir().'/.htbuilderstate';
		if (is_null($state)) {
			if (is_file($file)) unlink($file);
		} else {
			file_put_contents($file, $state);
		}
	}
	
	private function loadState() {
		$file = $this->getWebRootDir().'/.htbuilderstate';
		if (is_file($file)) {
			return intval(file_get_contents($file));
		} else {
			return 0;
		}
	}

	/** @return string */
	private function &getHomeDir() {
		if (is_null($this->homeDir)) {
			$dir = (isset($_SERVER['HOME']) && $_SERVER['HOME']) ? $_SERVER['HOME'] : ('/home/'.$this->user);
			if (!is_dir($dir) && is_dir(($altDir = '/home2/'.$this->user))) {
				$dir = $altDir;
			}
			$this->homeDir = $dir;
		}
		return $this->homeDir;
	}
	
	/**
	 * Get user's web root directory for current domain
	 * @return string
	 */
	private function &getWebRootDir() {
		if (!$this->webRootDir) {
			$homeDir = $this->getHomeDir();
			/** @return string */
			$buildRoot = function() use($homeDir) {
				$dir = $homeDir.'/domains/'.$this->domain.'/public_html';
				if (!is_dir($dir)) {
					if (!class_exists('\TrueBV\Punycode')) {
						require_once __DIR__.'/punycode/Exception/OutOfBoundsException.php';
						require_once __DIR__.'/punycode/Exception/DomainOutOfBoundsException.php';
						require_once __DIR__.'/punycode/Exception/LabelOutOfBoundsException.php';
						require_once __DIR__.'/punycode/Punycode.php';
					}
					$puny = new \TrueBV\Punycode();
					try {
						$encDomain = $puny->encode($this->domain);
						$dir = $homeDir.'/domains/'.$encDomain.'/public_html';
					} catch (Exception $ex) {}
				}
				return $dir;
			};
			$subdomain = $this->resolveSubdomain();
			if ($subdomain) {
				$fullSubdomain = $subdomain.'.'.$this->domain;
				$this->webRootDir = $homeDir.'/domains/'.$fullSubdomain.'/public_html';
				if (!is_dir($this->webRootDir)) {
					$this->webRootDir = $buildRoot().'/'.$subdomain;
				}
			} else {
				$this->webRootDir = $buildRoot();
			}
			if (!is_dir($this->webRootDir)) {
				throw new ErrorException("Path for domain not found");
			}
		}
		return $this->webRootDir;
	}

	public function extractFromUrl($url) {
		@set_time_limit(3600);
		$ch = curl_init();
		curl_setopt($ch, CURLOPT_URL, $url);
		curl_setopt($ch, CURLOPT_HTTPHEADER, array('Expect:'));
		curl_setopt($ch, CURLOPT_TIMEOUT, 3600);
		curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
		curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
		//curl_setopt($ch, CURLOPT_MAXREDIRS, 3);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		//curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
		curl_setopt($ch, CURLOPT_WRITEFUNCTION, array($this, 'readZip'));
		$r = curl_exec($ch);
		$status = curl_getinfo($ch, CURLINFO_HTTP_CODE);
		$error = curl_error($ch);
		if ($status != 200) {
			throw new ErrorException("Zip download error (code: $status)".($error ? ": $error" : ''));
		}
		curl_close($ch);
		if ($this->zipFile) { $this->zipFile->close(); }
	}
	
	/** @var ZipFile */
	private $zipFile = null;
	private $zipBuffer = "";
	
	protected function readZip($ch, $data) {
		$dir = $this->getWebRootDir();
		if (is_file($dir.'/index.html')) {
			rename($dir.'/index.html', $dir.'/index_backup.html');
		}
		$size = function_exists('mb_strlen') ? mb_strlen($data, '8bit') : strlen($data);
		$this->zipBuffer .= $data;
		//self::debugLog("Receive: ".$data);
		$continue = true;
		while ($continue) {
			$continue = false;
			$size_0 = function_exists('mb_strlen') ? mb_strlen($this->zipBuffer, '8bit') : strlen($this->zipBuffer);
			if ($this->zipFile && $size_0 >= $this->zipFile->compressedSize) {
				$continue = true;
				$this->zipFile->writeUncompressed(substr($this->zipBuffer, 0, $this->zipFile->compressedSize));
				$this->zipBuffer = substr($this->zipBuffer, $this->zipFile->compressedSize);
				$this->zipFile->close();
				$this->zipFile = null;
			} else if (!$this->zipFile && ($fp = strpos($this->zipBuffer, ZipFile::ZIP_FILE_HEAD)) !== false) {
				$continue = true;
				$this->zipBuffer = substr($this->zipBuffer, $fp);
				$hsize = 26 + strlen(ZipFile::ZIP_FILE_HEAD);
				if (strlen($this->zipBuffer) <= $hsize) { break; } // need more data
				$name_size_raw = unpack('v', substr($this->zipBuffer, strlen(ZipFile::ZIP_FILE_HEAD) + 22, 2));
				$name_size = intval(reset($name_size_raw));
				$extra_size_raw = unpack('v', substr($this->zipBuffer, strlen(ZipFile::ZIP_FILE_HEAD) + 24, 2));
				$extra_size = intval(reset($extra_size_raw));
				$hsize += $name_size + $extra_size;
				if (strlen($this->zipBuffer) < $hsize) { break; } // need more data
				// parse file header
				if ($this->zipFile) { $this->zipFile->close(); }
				$this->zipFile = ZipFile::parse(substr($this->zipBuffer, 0, $hsize), $dir);
				//self::debugLog('File: '.$this->zipFile->baseDir.'/'.$this->zipFile->name);
				$this->zipFile->open();
				$this->zipBuffer = substr($this->zipBuffer, $hsize);
			}
		}
		return $size;
	}
	
	private static function debugLog($msg) {
		$text = "[".date('Y-m-d H:i:s')."] $msg\n";
		file_put_contents('debug.log', $text, FILE_APPEND);
	}
}

class ZipFile {
	
	const ZIP_FILE_HEAD = "\x50\x4b\x03\x04";
	
	public $version;
	public $flags;
	public $compression;
	public $modDateTime;
	public $crc32;
	public $crc32Raw;
	public $compressedSize;
	public $size;
	public $sizeRaw;
	public $name;
	public $extra;
	public $baseDir;
	private $handle;
	
	public function __construct($baseDir = null) {
		$this->baseDir = rtrim($baseDir, '/');
		$this->modDateTime = date('Y-m-d H:i:s');
		$this->compressedSize = 0;
		$this->size = 0;
		$this->name = 'new file';
	}
	
	public function open() {
		$dir = dirname($this->baseDir.'/'.$this->name);
		if (!is_dir($dir)) { mkdir($dir, 0755, true); }
		if (!empty($this->name)) {
			$this->handle = fopen($this->baseDir.'/'.$this->name, 'w');
		} else {
			throw new ErrorException("File name is empty");
		}
	}
	
	public function close() {
		if ($this->handle !== false) {
			fclose($this->handle);
			chmod($this->baseDir.'/'.$this->name, 0644);
		}
	}
	
	public function write($data) {
		if ($this->handle !== false) {
			fwrite($this->handle, $data);
		} else {
			throw new ErrorException("File is not open");
		}
	}
	
	public function writeUncompressed($data) {
		$data_u = gzinflate($data);
		$this->write($data_u);
	}


	/**
	 * @param string $data
	 * @return ZipFile
	 */
	public static function parse($data, $baseDir = null) {
		$data = substr($data, strlen(self::ZIP_FILE_HEAD));
		$fh = new ZipFile($baseDir);
		$fh->version = substr($data, 0, 2);
		$fh->flags = substr($data, 2, 2);
		$fh->compression = substr($data, 4, 2);
		$hexdtime_raw = unpack('V', substr($data, 6, 4));
		$hexdtime = reset($hexdtime_raw);
		$fh->modDateTime = ''.((($hexdtime >> 25) & 0xff) + 1980).'-'.
				sprintf('%02d', ($hexdtime >> 21) & 0x0f).'-'.
				sprintf('%02d', ($hexdtime >> 16) & 0x1f).' '.
				sprintf('%02d', ($hexdtime >> 11) & 0x1f).':'.
				sprintf('%02d', ($hexdtime >>  5) & 0x3f).':'.
				sprintf('%02d', ($hexdtime <<  1) & 0x1f);
		$fh->crc32Raw = substr($data, 10, 4);
		$crc32 = unpack('V', $fh->crc32Raw);
		$fh->crc32 = sprintf('%08x', reset($crc32));
		$clen = unpack('V', substr($data, 14, 4));
		$fh->compressedSize = reset($clen);
		$fh->sizeRaw = substr($data, 18, 4);
		$ulen = unpack('V', $fh->sizeRaw);
		$fh->size = reset($ulen);
		$nlen_raw = unpack('v', substr($data, 22, 2));
		$nlen = reset($nlen_raw);
		$elen_raw = unpack('v', substr($data, 24, 2));
		$elen = reset($elen_raw);
		$fh->name = substr($data, 26, $nlen);
		$fh->extra = substr($data, 26 + $nlen, $elen);
		return $fh;
	}
}
